#pragma once
#include <Arduino.h>
#include <AceButton.h>

#include "board_definition.h"
#include "board_defaults.h"
#include "energy_meter.h"
#include "knx_params_cos.h"

/*
H8I8O board pin
*/

ace_button::AceButton *_buttons[NUM_IOS];
KnxEnergyMeter *_energyMeters[NUM_IOS];

uint32_t _recallValues[SCENES_IOS];
bool _sceneEnabled[SCENES_IOS];
bool _updateRecallValue[SCENES_IOS][4];
scene_type_t _sceneTypes[SCENES_IOS][4];

class KnxBoardIO
{
private:
    uint8_t index;
    io_mode_t mode = MODE_NONE;

    // EnergyMeter
    float lastEnergy = 0;
    typedef void (*energy_meter_callback_t)(KnxBoardIO *, float);
    energy_meter_callback_t energyMeterCallback;
    float callbackPerConsumed = DEFAULT_CALLBACK_PER_CONSUMED;
    bool energyMeterCyclic = false;
    uint16_t energyMeterCyclicDelay = DEFAULT_CYCLIC_DELAY;
    unsigned long energyMeterLastCyclicCallAt = 0;

public:
    uint32_t value = 0; // Value is public (it's easier to manipulate with)

    KnxBoardIO(uint8_t index_)
    {
        index = index_;
        _buttons[index] = nullptr;
        _energyMeters[index] = nullptr;

        if (supportsScenes())
        {
            _recallValues[index] = 0;
            _sceneEnabled[index] = false;
            for (uint8_t i = 0; i < 4; i++)
            {
                _sceneTypes[index][i] = SCENE_1BIT;
                _updateRecallValue[index][i] = true;
            }
        }
    }

    // Get index
    uint8_t getIndex() { return index; }

    // Get pin according to index of boart IO
    uint8_t getPin() { return ioPins[index]; }

    // Setup pin mode as input
    void setupAsInput() { pinMode(getPin(), INPUT_PULLUP); }

    // Setup pin mode as output
    void setupAsOutput() { pinMode(getPin(), OUTPUT); }

    bool read() { return digitalRead(getPin()); }

    void writeDigital(bool value) { digitalWrite(getPin(), value); }

    void writeAnalog(uint8_t value) { analogWrite(getPin(), value); }

    // Get IO mode
    io_mode_t getMode() { return mode; }

    // Set IO mode
    void setMode(io_mode_t mode_) { mode = mode_; }

    // Supports DC and LP ?

    bool supportsDcAndLp()
    {
        bool result = false;
        switch (mode)
        {
        case MODE_NONE:
        case MODE_IN_BUTTON:
        case MODE_IN_BUTTON_DC:
        case MODE_IN_BUTTON_LP:
            result = true;
        }
        return result;
    }

    /*
        Scenes
    */
    bool supportsScenes() { return index < SCENES_IOS; }

    // Get recall value pointer
    uint32_t *getRecallValuePtr()
    {
        if (supportsScenes())
        {
            return &_recallValues[index];
        }
        return nullptr;
    }

    // Get update recall value
    bool shouldUpdateRecallValue(uint8_t i)
    {
        if (supportsScenes())
        {
            return _updateRecallValue[index][i];
        }
        return false;
    }

    // Set update recall value
    void setShouldUpdateRecallValue(uint8_t i, bool value)
    {
        if (supportsScenes())
        {
            _updateRecallValue[index][i] = value;
        }
    }

    // Get scene enabled
    bool getSceneEnabled()
    {
        if (supportsScenes())
        {
            return _sceneEnabled[index];
        }
        return false;
    }

    // Toggle scene enabled value
    void toggleSceneEnabled()
    {
        if (supportsScenes())
        {
            _sceneEnabled[index] = !_sceneEnabled[index];
        }
    }

    // Get scene type
    scene_type_t getSceneType(uint8_t i)
    {
        if (supportsScenes())
        {
            return _sceneTypes[index][i];
        }
        return SCENE_1BIT;
    }

    // Set scene type
    void setSceneType(uint8_t i, scene_type_t type)
    {
        if (supportsScenes())
        {
            _sceneTypes[index][i] = type;
        }
    }

    /*
        AceButton section
    */

    // Setup AceButton object
    void setupAceButton(ace_button::ButtonConfig::EventHandler handler, bool defaultReleasedState = HIGH)
    {
        if (_buttons[index] == nullptr)
        {
            // Create AceButton
            _buttons[index] = new ace_button::AceButton(getPin(), defaultReleasedState, index);
            _buttons[index]->setEventHandler(handler);

            // Set some default configuration
            setAceButtonDebounceDelay(DEFAULT_DEBOUNCE);
            setAceButtonLongPressDelay(DEFAULT_LONG_PRESS_DELAY);
        }
    }

    // Set AceButton debounce delay
    void setAceButtonDebounceDelay(uint16_t delay) { _buttons[index]->getButtonConfig()->setDebounceDelay(delay); }

    // Set AceButton long press delay
    void setAceButtonLongPressDelay(uint16_t delay) { _buttons[index]->getButtonConfig()->setLongPressDelay(delay); }

    // Add click event to AceButton config
    void enableAceButtonClick() { _buttons[index]->getButtonConfig()->setFeature(ace_button::ButtonConfig::kFeatureClick); }

    // Add double click event to AceButton config
    void enableAceButtonDoubleClick() { _buttons[index]->getButtonConfig()->setFeature(ace_button::ButtonConfig::kFeatureDoubleClick); }

    // Add long press event to AceButton config
    void enableAceButtonLongPress() { _buttons[index]->getButtonConfig()->setFeature(ace_button::ButtonConfig::kFeatureLongPress); }

    /*
        EnergyMeter section
    */

    // Init energy meter
    void setupEnergyMeter(energy_meter_callback_t energyMeterCallback_, float pulsesPerKwh)
    {
        if (_energyMeters[index] == nullptr)
        {
            _energyMeters[index] = new KnxEnergyMeter(pulsesPerKwh);
            lastEnergy = _energyMeters[index]->getEnergy();
            energyMeterCallback = energyMeterCallback_;
        }
    }

    // Set energyMeterCallbackPerConsumed
    void setEnergyMeterCallbackPerConsumed(float value) { callbackPerConsumed = value; }

    // Set if callback should be called cyclicly
    void setEnergyMeterCyclic(bool enabled) { energyMeterCyclic = enabled; }

    // Set delay for cyclic sending
    void setEnergyMeterCyclicDelay(uint16_t delay) { energyMeterCyclicDelay = delay; }

    // Pulse energy meter
    void pulseEnergyMeter() { _energyMeters[index]->pulse(); }

    /*
        ETC section
    */

    // Loop
    void loop()
    {
        // Check button if initialised
        if (_buttons[index] != nullptr)
        {
            _buttons[index]->check();
        }

        // Check power meter if initialised
        if (_energyMeters[index] != nullptr)
        {
            float currentEnergy = _energyMeters[index]->getEnergy();

            if (energyMeterCyclic) // If cyclic, send in intervals
            {
                if (millis() - energyMeterLastCyclicCallAt >= energyMeterCyclicDelay)
                {
                    energyMeterCallback(this, currentEnergy);
                    energyMeterLastCyclicCallAt = millis();
                }
            }
            else // If per kWh, send per specified kWh(s)
            {
                if (currentEnergy - lastEnergy >= callbackPerConsumed)
                {
                    energyMeterCallback(this, currentEnergy);
                    lastEnergy = currentEnergy;
                }
            }
        }
    }
};
