#pragma once
#include <Arduino.h>
#include <AceButton.h>
#include <knx.h>

#include "board_io.h"
#include "board_ios.h"
#include "knx_actions.h"
#include "knx_utils.h"

void handleInputEvent(ace_button::AceButton *button, uint8_t event, uint8_t state)
{
    KnxBoardIO *io = getIo(button);

    // Handle MODE_IN_SWITCH
    if (io->getMode() == MODE_IN_SWITCH && (event == ace_button::AceButton::kEventPressed || event == ace_button::AceButton::kEventReleased))
    {
        actionSwitch(io, state);
    }
    // Handle MODE_IN_BUTTON
    else if ((io->getMode() == MODE_IN_SWITCH || io->getMode() == MODE_IN_BUTTON) && event == ace_button::AceButton::kEventClicked)
    {
        actionButton(io);
    }
    // Handle MODE_IN_BUTTON_DC
    else if (event == ace_button::AceButton::kEventDoubleClicked)
    {
        actionButtonDc(io);
    }
    // Handle MODE_IN_BUTTON_DC
    else if (event == ace_button::AceButton::kEventLongPressed)
    {
        actionButtonLp(io);
    }
    // Handle MODE_IN_ENERGY_METER
    else if (io->getMode() == MODE_IN_ENERGY_METER && event == ace_button::AceButton::kEventPressed)
    {
        io->pulseEnergyMeter();
    }
    else if (io->getMode() == MODE_SCENE_CONTROL && event == ace_button::AceButton::kEventClicked)
    {
        actionSceneIn(io);
    }
}

/* Since there is already problem with not enough EEPROM memory, this is unused */
void handleSave() { digitalWrite(LED_PIN, HIGH); }

void initConfiguration()
{
    // Board params from configuration
    bool rgbEnabled = getRgbEnabledParam();

    // Setup RGB if enabled
    if (rgbEnabled)
    {
        KnxRgbIos rgbIos = getRgbIos();
        rgbIos.rIo->setupAsOutput();
        rgbIos.gIo->setupAsOutput();
        rgbIos.bIo->setupAsOutput();

        getGroupObjectRgb().dataPointType(DPT_Colour_RGB);
        getGroupObjectRgb().callback(actionRGB);

        getGroupObjectRgbOnOff().dataPointType(DPT_Switch);
        getGroupObjectRgbOnOff().callback(actionRGBOnOff);
    }

    for (uint8_t i = 0; i < NUM_IOS; i++)
    {
        // Skip RGB pins if RGB is enabled
        if (rgbEnabled && (i == RGB_R_INDEX || i == RGB_G_INDEX || i == RGB_B_INDEX))
        {
            continue;
        }

        KnxBoardIO *io = getIo(i);

        // IO params from configuration
        io_mode_t mode = getModeParam(i);
        uint16_t debounce = getDebounceParam(i);
        init_state_t initState = getInitStateParam(i);
        switch_type_t switchType = getSwitchTypeParam(i);
        uint16_t emPulses = getEMPulsesParam(i);
        em_cyclic_t emCyclic = getEMCyclicParam(i);
        uint16_t emCbKwh = getEMCbKwmParam(i);
        uint16_t emCyclicDelay = getEMCyclicDelayParam(i);
        uint8_t initStateOut = getInitStateOutParam(i);
        uint8_t initStatePwmOut = getInitStatePwmOutParam(i);
        uint16_t lpDelay = getLpDelayParam(i);
        bool dcEnabled = getDcEnabledParam(i);
        bool lpEnabled = getLpEnabledParam(i);

        // Set mode to io
        io->setMode(mode);

        // Setup io by mode
        switch (mode)
        {
        case MODE_IN_SWITCH:
            io->setupAsInput();
            io->setupAceButton(handleInputEvent, switchType == TYPE_NORMALLY_OPEN ? HIGH : LOW);
            io->setAceButtonDebounceDelay(debounce);
            setNthByte(&io->value, 0, static_cast<uint8_t>(inputInitialValue(io, initState)));

            getGroupObjectIn(io->getIndex()).dataPointType(DPT_Switch);
            break;

        case MODE_IN_BUTTON:
            io->setupAsInput();
            io->setupAceButton(handleInputEvent, switchType == TYPE_NORMALLY_OPEN ? HIGH : LOW);
            io->enableAceButtonClick();
            io->setAceButtonDebounceDelay(debounce);
            setNthByte(&io->value, 0, static_cast<uint8_t>(inputInitialValue(io, initState)));

            getGroupObjectIn(io->getIndex()).dataPointType(DPT_Switch);
            break;

        case MODE_IN_ENERGY_METER:
            io->setupAsInput();
            io->setupAceButton(handleInputEvent, switchType == TYPE_NORMALLY_OPEN ? HIGH : LOW);
            io->setupEnergyMeter(actionEnergyMeter, emPulses);
            io->setAceButtonDebounceDelay(debounce);
            io->setEnergyMeterCallbackPerConsumed(emCbKwh);
            io->setEnergyMeterCyclicDelay(emCyclicDelay);
            io->setEnergyMeterCyclic(emCyclic == EM_SEND_CYCLIC);

            getGroupObjectInEm(io->getIndex()).dataPointType(DPT_ActiveEnergy_kWh);
            break;

        case MODE_OUT:
            io->setupAsOutput();
            io->value = initStateOut;
            io->writeDigital(io->value);

            getGroupObjectOut(io->getIndex()).dataPointType(DPT_Switch);
            getGroupObjectOut(io->getIndex()).callback(actionOut);
            break;

        case MODE_OUT_PWM:
            io->setupAsOutput();

            setNthByte(&io->value, 0, static_cast<uint8_t>(initStatePwmOut));
            setNthByte(&io->value, 1, static_cast<uint8_t>(initStateOut));

            io->writeAnalog(initStateOut ? initStatePwmOut : 0);

            getGroupObjectOut(io->getIndex()).dataPointType(DPT_Switch);
            getGroupObjectOut(io->getIndex()).callback(actionOutOnOffPwm);

            getGroupObjectOutPwm(io->getIndex()).dataPointType(DPT_Scaling);
            getGroupObjectOutPwm(io->getIndex()).callback(actionOutPwm);
            break;

        default:
            break;
        }

        if (mode == MODE_SCENE_CONTROL && io->supportsScenes())
        {
            scene_type_t sceneTypeA = getSceneTypeAParam(i);
            scene_type_t sceneTypeB = getSceneTypeBParam(i);
            scene_type_t sceneTypeC = getSceneTypeCParam(i);
            scene_type_t sceneTypeD = getSceneTypeDParam(i);

            bool sceneValue1BitA = getSceneValue1BitAParam(i);
            bool sceneValue1BitB = getSceneValue1BitBParam(i);
            bool sceneValue1BitC = getSceneValue1BitCParam(i);
            bool sceneValue1BitD = getSceneValue1BitDParam(i);

            uint8_t sceneValue8BitA = getSceneValue8BitAParam(i);
            uint8_t sceneValue8BitB = getSceneValue8BitBParam(i);
            uint8_t sceneValue8BitC = getSceneValue8BitCParam(i);
            uint8_t sceneValue8BitD = getSceneValue8BitDParam(i);

            scene_type_t sceneTypes[4] = {sceneTypeA, sceneTypeB, sceneTypeC, sceneTypeD};

            bool sceneValues1Bit[4] = {sceneValue1BitA, sceneValue1BitB, sceneValue1BitC, sceneValue1BitD};
            uint8_t sceneValues8Bit[4] = {sceneValue8BitA, sceneValue8BitB, sceneValue8BitC, sceneValue8BitD};

            void (*sceneActions1Bit[4])(GroupObject &) = {actionSceneA, actionSceneB, actionSceneC, actionSceneD};
            void (*sceneActions8Bit[4])(GroupObject &) = {actionScene8BitA, actionScene8BitB, actionScene8BitC, actionScene8BitD};

            io->setupAsInput();
            io->setupAceButton(handleInputEvent, switchType == TYPE_NORMALLY_OPEN ? HIGH : LOW);
            io->enableAceButtonClick();
            io->setAceButtonDebounceDelay(debounce);

            for (uint8_t j = 0; j < 4; j++)
            {
                io->setSceneType(j, sceneTypes[j]);

                if (sceneTypes[j] == SCENE_1BIT)
                {
                    setNthByte(&io->value, j, static_cast<uint8_t>(sceneValues1Bit[j]));
                    setNthByte(io->getRecallValuePtr(), j, 0);

                    getGroupObjectScene1Bit(i, j).dataPointType(DPT_Switch);
                    getGroupObjectScene1Bit(i, j).callback(sceneActions1Bit[j]);
                    getGroupObjectScene1Bit(i, j).requestObjectRead();
                }
                else
                {
                    setNthByte(&io->value, j, sceneValues8Bit[j]);
                    setNthByte(io->getRecallValuePtr(), j, 0);

                    getGroupObjectScene8Bit(i, j).dataPointType(DPT_Scaling);
                    getGroupObjectScene8Bit(i, j).callback(sceneActions8Bit[j]);
                    getGroupObjectScene8Bit(i, j).requestObjectRead();
                }
            }
        }

        if (io->supportsDcAndLp())
        {
            if (mode == MODE_IN_BUTTON_DC || dcEnabled)
            {
                io->setupAsInput();
                io->setupAceButton(handleInputEvent, switchType == TYPE_NORMALLY_OPEN ? HIGH : LOW);
                io->enableAceButtonDoubleClick();
                io->setAceButtonDebounceDelay(debounce);
                setNthByte(&io->value, 1, static_cast<uint8_t>(inputInitialValue(io, initState)));

                getGroupObjectInDc(io->getIndex()).dataPointType(DPT_Switch);
            }

            if (mode == MODE_IN_BUTTON_LP || lpEnabled)
            {
                io->setupAsInput();
                io->setupAceButton(handleInputEvent, switchType == TYPE_NORMALLY_OPEN ? HIGH : LOW);
                io->enableAceButtonLongPress();
                io->setAceButtonDebounceDelay(debounce);
                io->setAceButtonLongPressDelay(lpDelay);
                setNthByte(&io->value, 2, static_cast<uint8_t>(inputInitialValue(io, initState)));

                getGroupObjectInDc(io->getIndex()).dataPointType(DPT_Switch);
            }
        }
    }
}
