from models.base_model import *
from models.translation import Translation
from models.data_point_type import DataPointType, DataPointSubType

FlagType = Union[int, bool]


def convert_flag(flag: FlagType) -> int:
    flag: int = int(flag)
    return 1 if flag == 0 else 0


class ComObjectRef(BaseModel):
    type: Optional[str] = type_field('Kaenx.Creator.Models.ComObjectRef, Kaenx.Creator')

    uid: int = Field(alias='Uid')
    name: str = Field(alias='Name')
    com_object: int = Field(alias='ComObject')

    is_auto_generated: Optional[bool] = Field(alias='IsAutoGenerated', default=True)
    text: Optional[List[Translation]] = Field(alias='Text', default=Translation.from_string(str()))
    function_text: Optional[List[Translation]] = Field(alias='FunctionText', default=Translation.from_string(str()))

    translation_text: Optional[bool] = Field(alias='TranslationText', default=False)
    overwrite_text: Optional[bool] = Field(alias='OverwriteText', default=False)
    translation_function_text: Optional[bool] = Field(alias='TranslationFunctionText', default=False)
    overwrite_function_text: Optional[bool] = Field(alias='OverwriteFunctionText', default=False)
    overwrite_dpt: Optional[bool] = Field(alias='OverwriteDpt', default=False)
    overwrite_dpst: Optional[bool] = Field(alias='OverwriteDpst', default=False)
    sub_type_number: Optional[Any] = Field(alias='SubTypeNumber', default=None)
    type_number: Optional[Any] = Field(alias='TypeNumber', default=None)
    overwrite_o_s: Optional[bool] = Field(alias='OverwriteOS', default=False)
    object_size: Optional[int] = Field(alias='ObjectSize', default=1)
    flag_read: Optional[int] = Field(alias='FlagRead', default=1)
    overwrite_f_r: Optional[bool] = Field(alias='OverwriteFR', default=False)
    flag_write: Optional[int] = Field(alias='FlagWrite', default=1)
    overwrite_f_w: Optional[bool] = Field(alias='OverwriteFW', default=False)
    flag_trans: Optional[int] = Field(alias='FlagTrans', default=1)
    overwrite_f_t: Optional[bool] = Field(alias='OverwriteFT', default=False)
    flag_comm: Optional[int] = Field(alias='FlagComm', default=1)
    overwrite_f_c: Optional[bool] = Field(alias='OverwriteFC', default=False)
    flag_update: Optional[int] = Field(alias='FlagUpdate', default=1)
    overwrite_f_u: Optional[bool] = Field(alias='OverwriteFU', default=False)
    flag_on_init: Optional[int] = Field(alias='FlagOnInit', default=1)
    overwrite_f_o_i: Optional[bool] = Field(alias='OverwriteFOI', default=False)
    use_text_parameter: Optional[bool] = Field(alias='UseTextParameter', default=False)
    parameter_ref: Optional[int] = Field(alias='ParameterRef', default=-1)

    @computed_field(alias='Type', return_type=DataPointType)
    @property
    def data_point_type(self):
        return DataPointType.get_type_by_number(self.type_number)

    @computed_field(alias='SubType', return_type=DataPointSubType)
    @property
    def data_point_sub_type(self):
        return DataPointType.get_sub_type_by_number(self.type_number, self.sub_type_number)


class ComObject(BaseModel):
    type: Optional[str] = type_field('Kaenx.Creator.Models.ComObject, Kaenx.Creator')

    uid: int = Field(alias='Uid')
    name: str = Field(alias='Name')
    number: int = Field(alias='Number')
    type_number: str = Field(alias='TypeNumber')
    sub_type_number: str = Field(alias='SubTypeNumber')

    flag_read: int = Field(alias='FlagRead', default=0)
    flag_write: int = Field(alias='FlagWrite', default=0)
    flag_trans: int = Field(alias='FlagTrans', default=0)
    flag_comm: int = Field(alias='FlagComm', default=0)
    flag_update: int = Field(alias='FlagUpdate', default=0)
    flag_on_init: int = Field(alias='FlagOnInit', default=0)

    parameter_ref: Optional[int] = Field(alias='ParameterRef', default=-1)
    text: Optional[List[Translation]] = Field(alias='Text', default=Translation.from_string(str()))
    translation_text: Optional[bool] = Field(alias='TranslationText', default=False)
    function_text: Optional[List[Translation]] = Field(alias='FunctionText', default=Translation.from_string(str()))
    translation_function_text: Optional[bool] = Field(alias='TranslationFunctionText', default=False)
    type_value: Optional[Any] = Field(alias='TypeValue', default=None)
    has_dpt: Optional[bool] = Field(alias='HasDpt', default=True)
    has_dpts: Optional[bool] = Field(alias='HasDpts', default=True)
    object_size: Optional[int] = Field(alias='ObjectSize', default=1)
    use_text_parameter: Optional[bool] = Field(alias='UseTextParameter', default=False)

    @computed_field(alias='Type', return_type=DataPointType)
    @property
    def data_point_type(self):
        return DataPointType.get_type_by_number(self.type_number)

    @computed_field(alias='SubType', return_type=DataPointSubType)
    @property
    def data_point_sub_type(self):
        return DataPointType.get_sub_type_by_number(self.type_number, self.sub_type_number)

    @classmethod
    def from_args(cls, uid: int, name: str, type_number: str, sub_type_number: str,
                  text: Optional[str] = None, number: Optional[int] = None) -> 'ComObject':
        return cls.model_validate({
            'Uid': uid,
            'Name': name,
            'Number': number or uid,
            'TypeNumber': type_number,
            'SubTypeNumber': sub_type_number,
            'Text': Translation.from_string(text or name),
            'FunctionText': Translation.from_string(text or name)
        }, strict=False)

    def get_ref(self) -> ComObjectRef:
        ref_data = self.model_dump(exclude_unset=False, exclude_defaults=False, by_alias=True)
        ref_data.pop('$type')
        ref_data.update({'ComObject': self.uid})
        return ComObjectRef.model_validate(ref_data)

    def with_flags(self,
                   flag_comm: FlagType = 0, flag_read: FlagType = 0, flag_write: FlagType = 0,
                   flag_trans: FlagType = 0, flag_update: FlagType = 0, flag_on_init: FlagType = 0) -> 'ComObject':
        self.flag_read = convert_flag(flag_read)
        self.flag_write = convert_flag(flag_write)
        self.flag_trans = convert_flag(flag_trans)
        self.flag_comm = convert_flag(flag_comm)
        self.flag_update = convert_flag(flag_update)
        self.flag_on_init = convert_flag(flag_on_init)
        return self
